# Import Python Libraries here
import win32com.client


class Simulation:
    """The Simulation object represents CANoe's measurement functions in the Simulation mode.
    With the help of the Simulation object you can control the system time from an external source during the measurement.
    """

    def __init__(self, app) -> None:
        self.app = app
        self.log = self.app.log
        self.sim_obj = win32com.client.Dispatch(self.app.app_com_obj.Simulation)
        win32com.client.WithEvents(self.meas_obj, CanoeSimulationEvents)

    @property
    def animation(self) -> int:
        """Returns the animation factor.

        Returns:
            int: The animation factor.
        """
        return self.sim_obj.Animation

    @animation.setter
    def animation(self, value: int) -> None:
        """Sets the animation factor.

        Args:
            value (int): The animation factor.
        """
        self.sim_obj.Animation = value
        self.log.info(f'animation factor set to = {value}.')

    @property
    def current_time(self) -> int:
        """Returns the low-order 32 bit of the current system time.

        Returns:
            int: The low-order 32 bit of the current system time.
        """
        return self.sim_obj.CurrentTime

    @property
    def current_time_high(self) -> int:
        """Returns the high-order 32 bit of the current system time.

        Returns:
            int: The high-order 32 bit of the current system time.
        """
        return self.sim_obj.CurrentTimeHigh

    @property
    def notification_type(self) -> int:
        """gets the notification type for the OnIdle handler.

        Returns:
            int: The notification type. 0-Idle. 1-IdleU
        """
        return self.sim_obj.NotificationType

    @notification_type.setter
    def notification_type(self, value: int) -> None:
        """sets the notification type for the OnIdle handler.

        Args:
            value (int): The notification type. 0-Idle. 1-IdleU
        """
        self.sim_obj.NotificationType = value
        self.log.info(f'notification type set to = {value}.')

    def increment_time(self, ticks: int) -> None:
        """Increases the system time during the measurement in simulation mode.
        Whereas IncrementTime returns immediately after calling, the IncrementTimeAndWait function returns not until the simulation step is done in CANoe.

        Args:
            ticks (int): The number of ticks by which the system time is to be increased. 1 tick corresponds to 10 µs.
        """
        self.sim_obj.IncrementTime(ticks)
        self.log.info(f'Increased the system time to = {ticks} ticks.')

    def increment_time_and_wait(self, ticks: int) -> None:
        """Increases the system time during the measurement in Simulation mode.
        Whereas IncrementTime returns immediately after calling, the IncrementTimeAndWait method returns not until the simulation step is done in CANoe.

        Args:
            ticks (int): The number of ticks by which the system time is to be increased. 1 tick corresponds to 10 µs.
        """
        self.sim_obj.IncrementTimeAndWait(ticks)
        self.log.info(f'Increased the system time to = {ticks} ticks.')


class CanoeSimulationEvents:
    """Handler for CANoe Simulation events"""

    @staticmethod
    def OnIdle(time_high: int, time: int) -> None:
        """Occurs after a simulation step.

        Args:
            time_high (int): High-order 32 bit of the current simulation time.
            time (int): Low-order 32 bit of the current simulation time.
        """
        print(f'Simulation OnIdle event triggered. time_high = {time_high} and time = {time}')

    @staticmethod
    def OnIdleU(time_high: int, time: int) -> None:
        """Occurs after a simulation step.

        Args:
            time_high (int): High-order 32 bit of the current simulation time.
            time (int): Low-order 32 bit of the current simulation time.
        """
        print(f'Simulation OnIdleU event triggered. time_high = {time_high} and time = {time}')



"""Python package for controlling Vector CANoe tool"""

# Import Python Libraries here
import os
import pythoncom
import win32com.client
from typing import Union
from time import sleep as wait

# import CANoe utils here
from py_canoe_utils.py_canoe_logger import PyCanoeLogger
from py_canoe_utils.application import Application
from py_canoe_utils.bus import Bus, Signal
from py_canoe_utils.capl import Capl
from py_canoe_utils.configuration import Configuration
from py_canoe_utils.measurement import Measurement
from py_canoe_utils.networks import Networks
from py_canoe_utils.system import System, Namespaces, Variables
from py_canoe_utils.ui import Ui
from py_canoe_utils.version import Version


class CANoe:
    r"""The CANoe class represents the CANoe application.
    The CANoe class is the foundation for the object hierarchy.
    You can reach all other methods from the CANoe class instance.

    Examples:
        >>> # Example to open CANoe configuration, start measurement, stop measurement and close configuration.
        >>> canoe_inst = CANoe(py_canoe_log_dir=r'D:\.py_canoe')
        >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
        >>> canoe_inst.start_measurement()
        >>> wait(10)
        >>> canoe_inst.stop_measurement()
        >>> canoe_inst.quit()
    """

    def __init__(self, py_canoe_log_dir='', user_capl_functions=tuple()):
        """
        Args:
            py_canoe_log_dir (str): directory to store py_canoe log. example 'D:\\.py_canoe'
            user_capl_functions (tuple): user defined CAPL funtions to access.
        """
        pcl = PyCanoeLogger(py_canoe_log_dir)
        self.log = pcl.log
        self.app = Application(self.log)
        self.meas = object
        self.__diag_devices = dict()
        self.user_capl_function_names = user_capl_functions

    def open(self, canoe_cfg: str, visible=True, auto_save=False, prompt_user=False) -> None:
        r"""Loads CANoe configuration.

        Args:
            canoe_cfg (str): The complete path for the CANoe configuration.
            visible (bool): True if you want to see CANoe UI. Defaults to True.
            auto_save (bool, optional): A boolean value that indicates whether the active configuration should be saved if it has been changed. Defaults to False.
            prompt_user (bool, optional): A boolean value that indicates whether the user should intervene in error situations. Defaults to False.

        Examples:
            >>> # The following example opens a configuration
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
        """
        pythoncom.CoInitialize()
        self.app.app_com_obj = win32com.client.Dispatch('CANoe.Application')
        cav = self.app.app_com_obj.Version
        self.log.info(f'Dispatched Vector CANoe Application {cav.major}.{cav.minor}.{cav.Build}')
        self.app.app_com_obj.Configuration.Modified = False
        self.app.visible = visible
        self.app.open(path=canoe_cfg, auto_save=auto_save, prompt_user=prompt_user)
        self.meas = Measurement(self.app, self.user_capl_function_names)
        networks_obj = Networks(self.app)
        self.__diag_devices = networks_obj.fetch_diag_devices()

    def new(self, auto_save=False, prompt_user=False) -> None:
        """Creates a new configuration.

        Args:
            auto_save (bool, optional): A boolean value that indicates whether the active configuration should be saved if it has been changed. Defaults to False.
            prompt_user (bool, optional): A boolean value that indicates whether the user should intervene in error situations. Defaults to False.

        Examples:
            >>> # The following example creates a new configuration
            >>> canoe_inst = CANoe()
            >>> canoe_inst.new()
        """
        self.app.new(auto_save, prompt_user)

    def quit(self):
        r"""Quits CANoe without saving changes in the configuration.

        Examples:
            >>> # The following example quits CANoe
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.quit()
        """
        self.app.quit()

    def start_measurement(self) -> bool:
        r"""Starts the measurement.

        Returns:
            True if measurement started. else Flase.

        Examples:
            >>> # The following example starts the measurement
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
        """
        return self.meas.start()

    def stop_measurement(self) -> bool:
        r"""Stops the measurement.

        Returns:
            True if measurement stopped. else Flase.

        Examples:
            >>> # The following example stops the measurement
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.stop_measurement()
        """
        return self.meas.stop()

    def stop_ex_measurement(self) -> bool:
        r"""StopEx repairs differences in the behavior of the Stop method on deferred stops concerning simulated and real mode in CANoe.

        Returns:
            True if measurement stopped. else Flase.

        Examples:
            >>> # The following example full stops the measurement
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.stop_ex_measurement()
        """
        return self.meas.stop_ex()

    def reset_measurement(self) -> bool:
        r"""reset the measurement.

        Returns:
            Measurement running status(True/False).

        Examples:
            >>> # The following example resets the measurement
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.reset_measurement()
        """
        if self.meas.running:
            self.meas.stop()
        self.meas.start()
        self.log.info(f'Resetted measurement.')
        return self.meas.running

    def start_measurement_in_animation_mode(self, animation_delay=100) -> None:
        r"""Starts the measurement in Animation mode.

        Args:
            animation_delay (int): The animation delay during the measurement in Offline Mode.

        Examples:
            >>> # The following example starts the measurement in Animation mode
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement_in_animation_mode()
        """
        self.meas.animation_delay = animation_delay
        self.meas.animate()

    def break_measurement_in_offline_mode(self) -> None:
        r"""Interrupts the playback in Offline mode.

        Examples:
            >>> # The following example interrupts the playback in Offline mode
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.break_measurement_in_offline_mode()
        """
        self.meas.break_offline_mode()

    def reset_measurement_in_offline_mode(self) -> None:
        r"""Resets the measurement in Offline mode.

        Examples:
            >>> # The following example resets the measurement in Offline mode
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.reset_measurement_in_offline_mode()
        """
        self.meas.reset_offline_mode()

    def step_measurement_event_in_single_step(self) -> None:
        r"""Processes a measurement event in single step.

        Examples:
            >>> # The following example processes a measurement event in single step
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.step_measurement_event_in_single_step()
        """
        self.meas.step()

    def get_measurement_index(self) -> int:
        r"""gets the measurement index for the next measurement.

        Returns:
            Measurement Index.

        Examples:
            >>> # The following example gets the measurement index measurement
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.stop_measurement()
            >>> canoe_inst.get_measurement_index()
        """
        self.log.info(f'measurement_index value = {self.meas.measurement_index}')
        return self.meas.measurement_index

    def set_measurement_index(self, index: int) -> int:
        r"""sets the measurement index for the next measurement.

        Args:
            index (int): index value to set.

        Returns:
            Measurement Index value.

        Examples:
            >>> # The following example sets the measurement index for the next measurement to 15
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.stop_measurement()
            >>> canoe_inst.set_measurement_index(15)
        """
        self.meas.measurement_index = index
        return self.meas.measurement_index

    def get_measurement_running_status(self) -> bool:
        r"""Returns the running state of the measurement.

        Returns:
            True if The measurement is running.
            False if The measurement is not running.

        Examples:
            >>> # The following example returns measurement running status (True/False)
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.get_measurement_running_status()
        """
        self.log.info(f'CANoe Measurement Running Status = {self.meas.running}')
        return self.meas.running

    def save_configuration(self) -> bool:
        r"""Saves the configuration.

        Returns:
            True if configuration saved. else False.

        Examples:
            >>> # The following example saves the configuration if necessary
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.save_configuration()
        """
        conf_obj = Configuration(self.app)
        return conf_obj.save()

    def save_configuration_as(self, path: str, major: int, minor: int, create_dir=True) -> bool:
        r"""Saves the configuration as a different CANoe version.

        Args:
            path (str): The complete file name.
            major (int): The major version number of the target version.
            minor (int): The minor version number of the target version.
            create_dir (bool): create dirrectory if not available. default value True.

        Returns:
            True if configuration saved. else False.

        Examples:
            >>> # The following example saves the configuration as a CANoe 10.0 version
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.save_configuration_as(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo_v12.cfg', 10, 0)"""
        conf_obj = Configuration(self.app)
        config_path = '\\'.join(path.split('\\')[:-1])
        if not os.path.exists(config_path) and create_dir:
            os.makedirs(config_path, exist_ok=True)
        if os.path.exists(config_path):
            conf_obj.save_as(path, major, minor, False)
            return conf_obj.saved
        else:
            self.log.info(f'tried creating {path}. but {config_path} directory not found.')
            return False

    def get_signal_value(self, bus: str, channel: int, message: str, signal: str, raw_value=False) -> Union[float, int]:
        r"""get_signal_value Returns a Signal value.

        Args:
            bus (str): The Bus(CAN, LIN, FlexRay, MOST, AFDX, Ethernet) on which the signal is sent.
            channel (int): The channel on which the signal is sent.
            message (str): The name of the message to which the signal belongs.
            signal (str): The name of the signal.
            raw_value (bool): return raw value of the signal if true. Default(False) is physical value.

        Returns:
            signal vaue.

        Examples:
            >>> # The following example gets signal value
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> sig_val = canoe_inst.get_signal_value('CAN', 1, 'LightState', 'FlashLight')
            >>> print(sig_val)
        """
        bus_obj = Bus(self.app, bus_type=bus)
        sig_obj = Signal(bus_obj.get_signal(channel, message, signal))
        signal_value = sig_obj.raw_value if raw_value else sig_obj.value
        self.log.info(f'value of signal({bus}{channel}.{message}.{signal})={signal_value}.')
        return signal_value

    def set_signal_value(self, bus: str, channel: int, message: str, signal: str, value: int, raw_value=False) -> None:
        r"""set_signal_value sets a value to Signal. Works only when messages are sent using CANoe IL.

        Args:
            bus (str): The Bus(CAN, LIN, FlexRay, MOST, AFDX, Ethernet) on which the signal is sent.
            channel (int): The channel on which the signal is sent.
            message (str): The name of the message to which the signal belongs.
            signal (str): The name of the signal.
            value (Union[float, int]): signal value.
            raw_value (bool): return raw value of the signal if true. Default(False) is physical value.

        Examples:
            >>> # The following example sets signal value to 1
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.set_signal_value('CAN', 1, 'LightState', 'FlashLight', 1)
        """
        bus_obj = Bus(self.app, bus_type=bus)
        sig_obj = Signal(bus_obj.get_signal(channel, message, signal))
        if raw_value:
            sig_obj.raw_value = value
        else:
            sig_obj.value = value
        self.log.info(f'signal({bus}{channel}.{message}.{signal}) value set to {value}.')

    def get_signal_full_name(self, bus: str, channel: int, message: str, signal: str) -> str:
        """Determines the fully qualified name of a signal.

        Args:
            bus (str): The Bus(CAN, LIN, FlexRay, MOST, AFDX, Ethernet) on which the signal is sent.
            channel (int): The channel on which the signal is sent.
            message (str): The name of the message to which the signal belongs.
            signal (str): The name of the signal.

        Returns:
            str: The fully qualified name of a signal. The following format will be used for signals: <DatabaseName>::<MessageName>::<SignalName>
        """
        bus_obj = Bus(self.app, bus_type=bus)
        sig_obj = Signal(bus_obj.get_signal(channel, message, signal))
        return sig_obj.full_name

    def check_signal_online(self, bus: str, channel: int, message: str, signal: str) -> bool:
        r"""Checks whether the measurement is running and the signal has been received.

        Args:
            bus (str): The Bus(CAN, LIN, FlexRay, MOST, AFDX, Ethernet) on which the signal is sent.
            channel (int): The channel on which the signal is sent.
            message (str): The name of the message to which the signal belongs.
            signal (str): The name of the signal.

        Returns:
            TRUE if the measurement is running and the signal has been received. FALSE if not.

        Examples:
            >>> # The following example checks signal is online.
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.check_signal_online('CAN', 1, 'LightState', 'FlashLight')
        """
        bus_obj = Bus(self.app, bus_type=bus)
        sig_obj = Signal(bus_obj.get_signal(channel, message, signal))
        sig_online_status = sig_obj.is_online
        self.log.info(f'signal({bus}{channel}.{message}.{signal}) online status = {sig_online_status}.')
        return sig_online_status

    def check_signal_state(self, bus: str, channel: int, message: str, signal: str) -> int:
        r"""Checks whether the measurement is running and the signal has been received.

        Args:
            bus (str): The Bus(CAN, LIN, FlexRay, MOST, AFDX, Ethernet) on which the signal is sent.
            channel (int): The channel on which the signal is sent.
            message (str): The name of the message to which the signal belongs.
            signal (str): The name of the signal.

        Returns:
            State of the signal.
            0 The default value of the signal is returned.
            1 The measurement is not running; the value set by the application is returned.
            2 The measurement is not running; the value of the last measurement is returned.
            3 The signal has been received in the current measurement; the current value is returned.

        Examples:
            >>> # The following example checks signal state.
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.check_signal_state('CAN', 1, 'LightState', 'FlashLight')
        """
        bus_obj = Bus(self.app, bus_type=bus)
        sig_obj = Signal(bus_obj.get_signal(channel, message, signal))
        sig_state = sig_obj.state
        self.log.info(f'signal({bus}{channel}.{message}.{signal}) state = {sig_state}.')
        return sig_state

    def get_j1939_signal_value(self, bus: str, channel: int, message: str, signal: str, source_addr: int,
                               dest_addr: int,
                               raw_value=False) -> Union[float, int]:
        r"""get_j1939_signal Returns a Signal object.

        Args:
            bus (str): The Bus(CAN, LIN, FlexRay, MOST, AFDX, Ethernet) on which the signal is sent.
            channel (int): The channel on which the signal is sent.
            message (str): The name of the message to which the signal belongs.
            signal (str): The name of the signal.
            source_addr (int): The source address of the ECU that sends the message.
            dest_addr (int): The destination address of the ECU that receives the message.
            raw_value (bool): return raw value of the signal if true. Default(False) is physical value.

        Returns:
            signal vaue.

        Examples:
            >>> # The following example gets j1939 signal value
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> sig_val = canoe_inst.get_j1939_signal_value('CAN', 1, 'LightState', 'FlashLight', 0, 1)
            >>> print(sig_val)
        """
        bus_obj = Bus(self.app, bus_type=bus)
        sig_obj = Signal(bus_obj.get_j1939_signal(channel, message, signal, source_addr, dest_addr))
        signal_value = sig_obj.raw_value if raw_value else sig_obj.value
        self.log.info(f'value of signal({bus}{channel}.{message}.{signal})={signal_value}.')
        return signal_value

    def set_j1939_signal_value(self, bus: str, channel: int, message: str, signal: str, source_addr: int,
                               dest_addr: int, value: Union[float, int],
                               raw_value=False) -> None:
        r"""get_j1939_signal Returns a Signal object.

        Args:
            bus (str): The Bus(CAN, LIN, FlexRay, MOST, AFDX, Ethernet) on which the signal is sent.
            channel (int): The channel on which the signal is sent.
            message (str): The name of the message to which the signal belongs.
            signal (str): The name of the signal.
            source_addr (int): The source address of the ECU that sends the message.
            dest_addr (int): The destination address of the ECU that receives the message.
            value (Union[float, int]): signal value.
            raw_value (bool): return raw value of the signal if true. Default(False) is physical value.

        Returns:
            signal vaue.

        Examples:
            >>> # The following example gets j1939 signal value
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.set_j1939_signal_value('CAN', 1, 'LightState', 'FlashLight', 0, 1, 1)
        """
        bus_obj = Bus(self.app, bus_type=bus)
        sig_obj = Signal(bus_obj.get_j1939_signal(channel, message, signal, source_addr, dest_addr))
        if raw_value:
            sig_obj.raw_value = value
        else:
            sig_obj.value = value
        self.log.info(f'signal value set to {value}.')
        self.log.info(f'signal({bus}{channel}.{message}.{signal}) value set to {value}.')

    def get_j1939_signal_full_name(self, bus: str, channel: int, message: str, signal: str, source_addr: int,
                                   dest_addr: int) -> str:
        """Determines the fully qualified name of a signal.

        Args:
            bus (str): The Bus(CAN, LIN, FlexRay, MOST, AFDX, Ethernet) on which the signal is sent.
            channel (int): The channel on which the signal is sent.
            message (str): The name of the message to which the signal belongs.
            signal (str): The name of the signal.
            source_addr (int): The source address of the ECU that sends the message.
            dest_addr (int): The destination address of the ECU that receives the message.

        Returns:
            str: The fully qualified name of a signal. The following format will be used for signals: <DatabaseName>::<MessageName>::<SignalName>
        """
        bus_obj = Bus(self.app, bus_type=bus)
        sig_obj = Signal(bus_obj.get_j1939_signal(channel, message, signal, source_addr, dest_addr))
        return sig_obj.full_name

    def check_j1939_signal_online(self, bus: str, channel: int, message: str, signal: str, source_addr: int,
                                  dest_addr: int) -> bool:
        """Checks whether the measurement is running and the signal has been received.

        Args:
            bus (str): The Bus(CAN, LIN, FlexRay, MOST, AFDX, Ethernet) on which the signal is sent.
            channel (int): The channel on which the signal is sent.
            message (str): The name of the message to which the signal belongs.
            signal (str): The name of the signal.
            source_addr (int): The source address of the ECU that sends the message.
            dest_addr (int): The destination address of the ECU that receives the message.
        
        Returns:
            bool: TRUE: if the measurement is running and the signal has been received. FALSE: if not.
        """
        bus_obj = Bus(self.app, bus_type=bus)
        sig_obj = Signal(bus_obj.get_j1939_signal(channel, message, signal, source_addr, dest_addr))
        return sig_obj.is_online

    def check_j1939_signal_state(self, bus: str, channel: int, message: str, signal: str, source_addr: int,
                                 dest_addr: int) -> int:
        """Returns the state of the signal.

        Returns:
            int: State of the signal; possible values are: 0: The default value of the signal is returned. 1: The measurement is not running; the value set by the application is returned. 3: The signal has been received in the current measurement; the current value is returned.
        """
        bus_obj = Bus(self.app, bus_type=bus)
        sig_obj = Signal(bus_obj.get_j1939_signal(channel, message, signal, source_addr, dest_addr))
        return sig_obj.state

    def ui_activate_desktop(self, name: str) -> None:
        r"""Activates the desktop with the given name.

        Args:
            name (str): The name of the desktop to be activated.

        Examples:
            >>> # The following example switches to the desktop with the name "Configuration"
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.ui_activate_desktop("Configuration")
        """
        ui_obj = Ui(self.app)
        ui_obj.activate_desktop(name)

    def ui_open_baudrate_dialog(self) -> None:
        r"""opens the dialog for configuring the bus parameters. Make sure Measurement stopped when using this method.

        Examples:
            >>> # The following example opens the dialog for configuring the bus parameters
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.stop_measurement()
            >>> canoe_inst.ui_open_baudrate_dialog()
        """
        ui_obj = Ui(self.app)
        ui_obj.open_baudrate_dialog()

    def write_text_in_write_window(self, text: str) -> None:
        r"""Outputs a line of text in the Write Window.
        Args:
            text (str): The text.

        Examples:
            >>> # The following example Outputs a line of text in the Write Window.
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> wait(1)
            >>> canoe_inst.write_text_in_write_window("hello from python!")
            >>> wait(1)
            >>> print(canoe_inst.read_text_from_write_window())
        """
        ui_obj = Ui(self.app)
        ui_obj.send_text_to_write_window(text)

    def read_text_from_write_window(self) -> str:
        r"""read the text contents from Write Window.

        Returns:
            The text content.

        Examples:
            >>> # The following example reads text from Write Window.
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> wait(1)
            >>> canoe_inst.write_text_in_write_window("hello from python!")
            >>> wait(1)
            >>> print(canoe_inst.read_text_from_write_window())
        """
        ui_obj = Ui(self.app)
        return ui_obj.get_write_window_text_content()

    def clear_write_window_content(self) -> None:
        r"""Clears the contents of the Write Window.

        Examples:
            >>> # The following example clears content from Write Window.
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> wait(1)
            >>> canoe_inst.write_text_in_write_window("hello from python!")
            >>> wait(1)
            >>> canoe_inst.clear_write_window_content()
        """
        ui_obj = Ui(self.app)
        ui_obj.clear_write_window_content()

    def enable_write_window_output_file(self, output_file: str) -> None:
        r"""Enables logging of all outputs of the Write Window in the output file.

        Args:
            output_file (str): The complete path of the output file.

        Examples:
            >>> # The following example Enables logging of all outputs of the Write Window in the output file.
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.enable_write_window_output_file(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\write_out.txt')
            >>> canoe_inst.start_measurement()
            >>> wait(1)
            >>> canoe_inst.write_text_in_write_window("hello from python!")
            >>> wait(1)
            >>> canoe_inst.stop_measurement()
        """
        ui_obj = Ui(self.app)
        ui_obj.enable_write_window_logging(output_file)

    def disable_write_window_output_file(self) -> None:
        r"""Disables logging of all outputs of the Write Window.

        Examples:
            >>> # The following example Disables logging of all outputs of the Write Window.
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.disable_write_window_output_file()
        """
        ui_obj = Ui(self.app)
        ui_obj.disable_write_window_logging()

    def get_can_bus_statistics(self, channel: int) -> dict:
        r"""Returns CAN Bus Statistics.

        Args:
            channel (int): The channel of the statistic that is to be returned.

        Returns:
            CAN bus statistics.

        Examples:
            >>> # The following example prints CAN channel 1 statistics
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> print(canoe_inst.get_can_bus_statistics(channel=1))
        """
        conf_obj = Configuration(self.app)
        bus_types = {'CAN': 1, 'J1939': 2, 'TTP': 4, 'LIN': 5, 'MOST': 6, 'Kline': 14}
        bus_statistics_obj = conf_obj.conf_com_obj.OnlineSetup.BusStatistics.BusStatistic(bus_types['CAN'], channel)
        statistics_info = {
            # The bus load
            'bus_load': bus_statistics_obj.BusLoad,
            # The controller status
            'chip_state': bus_statistics_obj.ChipState,
            # The number of Error Frames per second
            'error': bus_statistics_obj.Error,
            # The total number of Error Frames
            'error_total': bus_statistics_obj.ErrorTotal,
            # The number of messages with extended identifier per second
            'extended': bus_statistics_obj.Extended,
            # The total number of messages with extended identifier
            'extended': bus_statistics_obj.ExtendedTotal,
            # The number of remote messages with extended identifier per second
            'extended_remote': bus_statistics_obj.ExtendedRemote,
            # The total number of remote messages with extended identifier
            'extended_remote_total': bus_statistics_obj.ExtendedRemoteTotal,
            # The number of overload frames per second
            'overload': bus_statistics_obj.Overload,
            # The total number of overload frames
            'overload_total': bus_statistics_obj.OverloadTotal,
            # The maximum bus load in 0.01 %
            'peak_load': bus_statistics_obj.PeakLoad,
            # Returns the current number of the Rx error counter
            'rx_error_count': bus_statistics_obj.RxErrorCount,
            # The number of messages with standard identifier per second
            'standard': bus_statistics_obj.Standard,
            # The total number of remote messages with standard identifier
            'standard_total': bus_statistics_obj.StandardTotal,
            # The number of remote messages with standard identifier per second
            'standard_remote': bus_statistics_obj.StandardRemote,
            # The total number of remote messages with standard identifier
            'standard_remote_total': bus_statistics_obj.StandardRemoteTotal,
            # The current number of the Tx error counter
            'tx_error_count': bus_statistics_obj.TxErrorCount,
        }
        self.log.info(f'CAN Bus Statistics: {statistics_info}.')
        return statistics_info

    def get_canoe_version_info(self) -> dict:
        r"""The Version class represents the version of the CANoe application.

        Returns:
            "full_name" - The complete CANoe version.
            "name" - The CANoe version.
            "build" - The build number of the CANoe application.
            "major" - The major version number of the CANoe application.
            "minor" - The minor version number of the CANoe application.
            "patch" - The patch number of the CANoe application.

        Examples:
            >>> # The following example returns CANoe application version relevant information.
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_version_info = canoe_inst.get_canoe_version_info()
            >>> print(canoe_version_info)
        """
        ver_obj = Version(self.app)
        version_info = {'full_name': ver_obj.full_name,
                        'name': ver_obj.name,
                        'build': ver_obj.build,
                        'major': ver_obj.major,
                        'minor': ver_obj.minor,
                        'patch': ver_obj.patch}
        self.log.info('> CANoe Application.Version <'.center(100, '='))
        for k, v in version_info.items():
            self.log.info(f'{k:<10}: {v}')
        self.log.info(''.center(100, '='))
        return version_info

    def define_system_variable(self, sys_var_name: str, value=0) -> object:
        r"""define_system_variable Create a system variable with an initial value
        Args:
            sys_var_name (str): The name of the system variable. Ex- "sys_var_demo::speed"
            value (Union[int, float, str]): variable value. Default value 0.
        
        Returns:
            object: The new Variable object.
        
        Examples:
            >>> # The following example gets system variable value
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.define_system_variable('sys_demo::speed', 1)
        """
        namespace = '::'.join(sys_var_name.split('::')[:-1])
        variable_name = sys_var_name.split('::')[-1]
        new_var_obj = None
        try:
            sys_obj = System(self.app)
            namespaces_obj = Namespaces(win32com.client.Dispatch(sys_obj.sys_com_obj.Namespaces))
            new_namespace_obj = namespaces_obj.add(namespace)
            vars_obj = Variables(new_namespace_obj.Variables)
            new_var_obj = vars_obj.add(variable_name, value)
            self.log.info(f'system variable({sys_var_name}) created and value set to {value}.')
        except Exception as e:
            self.log.info(f'failed to create system variable({sys_var_name}). {e}')
        return new_var_obj

    def get_system_variable_value(self, sys_var_name: str) -> Union[int, float, str, None]:
        r"""get_system_variable_value Returns a system variable value.

        Args:
            sys_var_name (str): The name of the system variable. Ex- "sys_var_demo::speed"

        Returns:
            System Variable value.

        Examples:
            >>> # The following example gets system variable value
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> sys_var_val = canoe_inst.get_system_variable_value('sys_var_demo::speed')
            >>>print(sys_var_val)
        """
        namespace = '::'.join(sys_var_name.split('::')[:-1])
        variable_name = sys_var_name.split('::')[-1]
        return_value = None
        try:
            sys_obj = System(self.app)
            namespace_object = sys_obj.sys_com_obj.Namespaces(namespace)
            return_value = namespace_object.Variables(variable_name).Value
            self.log.info(f'system variable({sys_var_name}) value = {return_value}.')
        except Exception as e:
            self.log.info(f'failed to get system variable({sys_var_name}) value. {e}')
        return return_value

    def set_system_variable_value(self, sys_var_name: str, value: Union[int, float, str]) -> None:
        r"""set_system_variable_value sets a value to system variable.

        Args:
            sys_var_name (str): The name of the system variable. Ex- "sys_var_demo::speed"
            value (Union[int, float, str]): variable value.

        Examples:
            >>> # The following example sets system variable value to 1
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.set_system_variable_value('sys_var_demo::speed', 1)
        """
        namespace = '::'.join(sys_var_name.split('::')[:-1])
        variable_name = sys_var_name.split('::')[-1]
        try:
            sys_obj = System(self.app)
            namespace_object = sys_obj.sys_com_obj.Namespaces(namespace)
            namespace_object.Variables(variable_name).Value = value
            self.log.info(f'system variable({sys_var_name}) value set to {value}.')
        except Exception as e:
            self.log.info(f'failed to set system variable({sys_var_name}) value. {e}')

    def send_diag_request(self, diag_ecu_qualifier_name: str, request: str, request_in_bytes=True,
                          return_sender_name=False) -> Union[str, dict]:
        r"""The send_diag_request method represents the query of a diagnostic tester (client) to an ECU (server) in CANoe.

        Args:
            diag_ecu_qualifier_name (str): Diagnostic Node ECU Qualifier Name configured in "Diagnostic/ISO TP Configuration".
            request (str): Diagnostic request in bytes or diagnostic request qualifier name.
            request_in_bytes (bool): True if Diagnostic request is bytes. False if you are using Qualifier name. Default is True.
            return_sender_name (bool): True if you user want response along with response sender name in dictionary. Default is False.

        Returns:
            diagnostic response stream. Ex- "50 01 00 00 00 00" or {'Door': "50 01 00 00 00 00"}

        Examples:
            >>> # Example 1 - The following example sends diagnostic request "10 01"
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> wait(1)
            >>> resp = canoe_inst.send_diag_request('Door', '10 01')
            >>> print(resp)
            >>> canoe_inst.stop_measurement()
            >>> # Example 2 - The following example sends diagnostic request "DefaultSession_Start"
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(canoe_cfg=r'C:\Users\Public\Documents\Vector\CANoe\Sample Configurations 11.0.81\.\CAN\Diagnostics\UDSBasic\UDSBasic.cfg')
            >>> canoe_inst.start_measurement()
            >>> wait(1)
            >>> resp = canoe_inst.send_diag_request('Door', 'DefaultSession_Start', False)
            >>> print(resp)
            >>> canoe_inst.stop_measurement()
        """
        diag_response_data = ""
        diag_responses = {}
        try:
            if diag_ecu_qualifier_name in self.__diag_devices.keys():
                self.log.info(f'{diag_ecu_qualifier_name}: Diagnostic Request --> {request}')
                if request_in_bytes:
                    diag_req_in_bytes = bytearray()
                    request = ''.join(request.split(' '))
                    for i in range(0, len(request), 2):
                        diag_req_in_bytes.append(int(request[i:i + 2], 16))
                    diag_req = self.__diag_devices[diag_ecu_qualifier_name].CreateRequestFromStream(diag_req_in_bytes)
                else:
                    diag_req = self.__diag_devices[diag_ecu_qualifier_name].CreateRequest(request)
                diag_req.Send()
                while diag_req.Pending:
                    wait(0.1)
                if diag_req.Responses.Count == 0:
                    self.log.info("Diagnostic Response Not Received.")
                else:
                    for k in range(1, diag_req.Responses.Count + 1):
                        diag_res = diag_req.Responses(k)
                        diag_response_data = " ".join(f"{d:02X}" for d in diag_res.Stream).upper()
                        diag_responses[diag_res.Sender] = diag_response_data
                        if diag_res.Positive:
                            self.log.info(f"{diag_res.Sender}: Diagnostic Response +ve <-- {diag_response_data}")
                        else:
                            self.log.info(f"{diag_res.Sender}: Diagnostic Response -ve <-- {diag_response_data}")
            else:
                self.log.info(f'Diagnostic ECU qualifier({diag_ecu_qualifier_name}) not available in loaded CANoe config.')
        except Exception as e:
            self.log.info(f'failed to send diagnostic request({request}). {e}')
        return diag_responses if return_sender_name else diag_response_data

    def __fetch_replay_blocks(self) -> dict:
        replay_blocks = dict()
        try:
            bus_obj = Bus(self.app)
            for replay_block in bus_obj.ReplayCollection:
                replay_blocks[replay_block.Name] = replay_block
        except Exception as e:
            self.log.info(f'failed to fetch replay blocks. {e}')
        return replay_blocks

    def set_replay_block_file(self, block_name: str, recording_file_path: str) -> None:
        r"""Method for setting CANoe replay block file.

        Args:
            block_name: CANoe replay block name
            recording_file_path: CANoe replay recording file including path.

        Examples:
            >>> # The following example sets replay block file
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.set_replay_block_file(block_name='replay block name', recording_file_path='replay file including path')
            >>> canoe_inst.start_measurement()
        """
        replay_blocks = self.__fetch_replay_blocks()
        if block_name in replay_blocks.keys():
            replay_blocks[block_name].Path = recording_file_path
            self.log.info(f'Replay block "{block_name}" updated with "{recording_file_path}" path.')
        else:
            self.log.warning(f'Replay block "{block_name}" not available.')

    def control_replay_block(self, block_name: str, start_stop: bool) -> None:
        r"""Method for setting CANoe replay block file.

        Args:
            block_name (str): CANoe replay block name
            start_stop (bool): True to start replay block. False to Stop.

        Examples:
            >>> # The following example starts replay block
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.set_replay_block_file(block_name='replay block name', recording_file_path='replay file including path')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.control_replay_block('replay block name', True)
        """
        replay_blocks = self.__fetch_replay_blocks()
        if block_name in replay_blocks.keys():
            if start_stop:
                replay_blocks[block_name].Start()
            else:
                replay_blocks[block_name].Stop()
            self.log.info(f'Replay block "{block_name}" {"Started" if start_stop else "Stopped"}.')
        else:
            self.log.warning(f'Replay block "{block_name}" not available.')

    def compile_all_capl_nodes(self) -> None:
        r"""compiles all CAPL, XML and .NET nodes.
        """
        capl_obj = Capl(self.app)
        capl_obj.compile()
        self.log.info(f'compiled all nodes successfully.')

    def call_capl_function(self, name: str, *arguments) -> bool:
        r"""Calls a CAPL function.
        Please note that the number of parameters must agree with that of the CAPL function.
        not possible to read return value of CAPL function at the moment. only execution status is returned.

        Args:
            name (str): The name of the CAPL function. Please make sure this name is already passed as argument during CANoe instance creation. see example for more info.
            arguments (tuple): Function parameters p1…p10 (optional).

        Returns:
            bool: CAPL function execution status. True-success, False-failed.

        Examples:
            >>> # The following example starts replay block
            >>> canoe_inst = CANoe(user_capl_functions=('addition_function', ))
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.start_measurement()
            >>> canoe_inst.call_capl_function('addition_function', 100, 200)
            >>> canoe_inst.stop_measurement()
        """
        capl_obj = Capl(self.app)
        exec_sts = capl_obj.call_capl_function(self.meas.user_capl_function_obj_dict[name], *arguments)
        self.log.info(f'triggered capl function({name}). execution status = {exec_sts}.')
        return exec_sts

    def add_offline_source_log_file(self, absolute_log_file_path: str) -> bool:
        """this method adds offline source log file.

        Args:
            absolute_log_file_path (str): absolute path of offline source log file.

        Returns:
            bool: returns True if log file added or already available. False if log file not available.
        
        Examples:
            >>> canoe_inst = CANoe()
            >>> canoe_inst.open(r'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\demo.cfg')
            >>> canoe_inst.add_offline_source_log_file(fr'D:\_kms_local\vector_canoe\py_canoe\demo_cfg\Logs\demo_log.blf')
        """
        if os.path.isfile(absolute_log_file_path):
            offline_sources = self.app.app_com_obj.Configuration.OfflineSetup.Source.Sources
            file_already_added = any([file == absolute_log_file_path for file in offline_sources])
            if file_already_added:
                self.log.info(f'offline logging file ({absolute_log_file_path}) already added.')
            else:
                offline_sources.Add(absolute_log_file_path)
                self.log.info(f'added offline logging file ({absolute_log_file_path})')
            return True
        else:
            self.log.info(f'invalid logging file ({absolute_log_file_path}). Failed to add.')
            return False



# Import Python Libraries here
import pythoncom
import win32com.client


class Networks:
    def __init__(self, app_obj) -> None:
        self.app_obj = app_obj
        self.log = self.app_obj.log
        self.networks_com_obj = win32com.client.Dispatch(self.app_obj.app_com_obj.Networks)

    def fetch_diag_devices(self) -> dict:
        diag_devices = {}
        for network in self.networks_com_obj:
            for device in network.Devices:
                try:
                    diag_devices[device.Name] = device.Diagnostic
                except pythoncom.com_error:
                    pass
        return diag_devices


class Network:
    def __init__(self, network_com_obj) -> None:
        self.network_com_obj = network_com_obj

    @property
    def bus_type(self) -> int:
        """Determines the bus type of the network

        Returns:
            int: The type of the network: 0-Invalid, 1-CAN, 2-J1939, 5-LIN, 6-MOST, 7-FlexRay, 9-J1708, 11-Ethernet
        """
        return self.network_com_obj.BusType

    @property
    def devices(self) -> object:
        """Returns the Devices object.

        Returns:
            object: The Devices object
        """
        return self.network_com_obj.Devices

    @property
    def name(self) -> str:
        return self.network_com_obj.Name



# Import Python Libraries here
import win32com.client


class Ui:
    """The UI object represents the user interface in CANoe.
    """

    def __init__(self, app) -> None:
        self.app = app
        self.log = self.app.log
        self.ui_com_obj = win32com.client.Dispatch(self.app.app_com_obj.UI)

    def get_command_availability(self, command: str) -> bool:
        """defines the availability of a command on the user interface.

        Args:
            command (str): The command. If no command is entered the function acts on all commands that can be influenced via this interface.

        Returns:
            bool: The availability of the command: If the command is available True is returned. Otherwise False is returned.
        """
        return self.ui_com_obj.CommandEnabled(command)

    def set_command_availability(self, command: str, value: bool) -> None:
        """sets the availability of a command on the user interface.

        Args:
            command (str): The command. If no command is entered the function acts on all commands that can be influenced via this interface.
            value (bool): A boolean value that indicates whether the command should be available. Possible values are: True: The command is available. False: The command is not available.
        """
        ce_obj = self.ui_com_obj.CommandEnabled(command)
        ce_obj = value
        self.log.info(f'enabled command {command}.')

    def activate_desktop(self, name: str) -> None:
        """Activates the desktop with the given name.

        Args:
            name (str): The name of the desktop to be activated.
        """
        self.ui_com_obj.ActivateDesktop(name)
        self.log.info(f'Activated the desktop with the given name({name}.')

    def open_baudrate_dialog(self) -> None:
        """Configures the bus parameters.
        """
        self.ui_com_obj.OpenBaudrateDialog()
        self.log.info(f'baudrate dialog opened. Configure the bus parameters.')

    def get_write_window_text_content(self) -> str:
        write_obj = Write(self)
        return write_obj.text

    def clear_write_window_content(self) -> None:
        write_obj = Write(self)
        write_obj.clear()

    def copy_write_window_content_to_clipboard(self) -> None:
        write_obj = Write(self)
        write_obj.copy()

    def disable_write_window_logging(self, tab_index=None) -> None:
        write_obj = Write(self)
        write_obj.disable_output_file(tab_index)

    def enable_write_window_logging(self, output_file: str, tab_index=None) -> None:
        write_obj = Write(self)
        write_obj.enable_output_file(output_file, tab_index)

    def send_text_to_write_window(self, text: str) -> None:
        write_obj = Write(self)
        write_obj.output(text)


class Write:
    """The Write object represents the Write Window in CANoe.
    It is part of the user interface.
    """

    def __init__(self, ui_obj) -> None:
        self.ui_obj = ui_obj
        self.log = self.ui_obj.log
        self.write_com_obj = win32com.client.Dispatch(self.ui_obj.ui_com_obj.Write)

    @property
    def text(self) -> str:
        """Gets the text contents of the Write window.

        Returns:
            str: The text content
        """
        return self.write_com_obj.Text

    def clear(self) -> None:
        """Clears the contents of the Write Window
        """
        self.write_com_obj.Clear()
        self.log.info(f'Cleared the contents of the Write Window.')

    def copy(self) -> None:
        """Copies the contents of the Write Window to the clipboard.
        """
        self.write_com_obj.Copy()
        self.log.info(f'Copied the contents of the Write Window to the clipboard.')

    def disable_output_file(self, tab_index=None) -> None:
        """Disables logging of all outputs of the Write Window for the certain page.

        Args:
            tab_index (int, optional): The index of the tab, for which logging of the output is to be deactivated. Defaults to None.
        """
        if tab_index:
            self.write_com_obj.DisableOutputFile(tab_index)
        else:
            self.write_com_obj.DisableOutputFile()
        self.log.info(f'Disabled logging of outputs of the Write Window. tab_index={tab_index}')

    def enable_output_file(self, output_file: str, tab_index=None) -> None:
        """Enables logging of all outputs of the Write Window in the output file for the certain page.

        Args:
            output_file (str, optional): The complete path of the output file. Defaults to None.
            tab_index (int, optional): The index of the page, for which logging of the output is to be activated. Defaults to None.
        """
        if tab_index:
            self.write_com_obj.EnableOutputFile(output_file, tab_index)
        else:
            self.write_com_obj.EnableOutputFile(output_file)
        self.log.info(
            f'Enabled logging of outputs of the Write Window. output_file={output_file} and tab_index={tab_index}')

    def output(self, text: str) -> None:
        """Outputs a line of text in the Write Window.

        Args:
            text (str): The text
        """
        self.write_com_obj.Output(text)
        self.log.info(f'Outputed {text} in the Write Window.')



# Import Python Libraries here


class Bus:
    """The Bus object represents a bus of the CANoe application.
    """

    def __init__(self, app_obj, bus_type='CAN') -> None:
        """Returns a Signal object.

        Args:
            app_obj (object): application class instance object.
            bus_type (str, optional): The desired bus type. Valid types are: CAN, LIN, FlexRay, AFDX, Ethernet. Defaults to 'CAN'.
        """
        self.app_obj = app_obj
        self.log = self.app_obj.log
        self.bus_com_obj = self.app_obj.app_com_obj.GetBus(bus_type)

    @property
    def active(self) -> bool:
        """determines the status of the Bus object.

        Returns:
            bool: The status of the Bus object.
        """
        return self.bus_com_obj.Active

    @active.setter
    def active(self, value: bool) -> None:
        """Sets the status of the Bus object.

        Args:
            value (bool): A boolean value that indicates whether the bus is to be simulated: TRUE: The bus will be simulated. FALSE: The bus will not be simulated.
        """
        self.bus_com_obj.Active = value
        self.log.info(f'status of the Bus object set to {value}.')

    def baudrate(self, channel_number: int) -> int:
        """Determines the baud rate of a channel.

        Args:
            channel_number (int): The channel number.

        Returns:
            int: The current baud rate of the channel.
        """
        return self.bus_com_obj.Baudrate(channel_number)

    @property
    def bus_name(self) -> str:
        """returns the bus name.

        Returns:
            str: The bus name.
        """
        return self.bus_com_obj.Name

    def set_bus_name(self, name: str) -> None:
        """Sets the bus name

        Args:
            name (str): The new name.
        """
        self.bus_com_obj.Name = name
        self.log.info(f'bus name set to {name}.')

    def get_signal(self, channel: int, message: str, signal: str) -> object:
        """Returns a Signal object.

        Args:
            channel (int): The channel on which the signal is sent.
            message (str): The name of the message to which the signal belongs.
            signal (str): The name of the signal.

        Returns:
            object: The Signal object.
        """
        return self.bus_com_obj.GetSignal(channel, message, signal)

    def get_j1939_signal(self, channel: int, message: str, signal: str, source_address: int,
                         destination_address: int) -> object:
        """Returns a Signal object.

        Args:
            channel (int): The channel on which the signal is sent.
            message (str): The name of the message to which the signal belongs.
            signal (str): The name of the signal.
            source_address (int): The source address of the ECU that sends the message
            destination_address (int): The destination address of the ECU that receives the message. For signals of global Parameter Groups (PDU 2 format) this parameter is not considered.

        Returns:
            object: The Signal object.
        """
        return self.bus_com_obj.GetJ1939Signal(channel, message, signal, source_address, destination_address)


class Channel:
    def __init__(self) -> None:
        pass


class Database:
    def __init__(self) -> None:
        pass


class Generator:
    def __init__(self) -> None:
        pass


class Node:
    def __init__(self) -> None:
        pass


class ReplayBlock:
    def __init__(self) -> None:
        pass


class Signal:
    """The Signal object represents a signal on the bus.
    """

    def __init__(self, signal_com_object) -> None:
        self.sig_com_obj = signal_com_object

    @property
    def full_name(self) -> str:
        """Determines the fully qualified name of a signal.

        Returns:
            str: The fully qualified name of a signal or a message. The following format will be used for signals: <DatabaseName>::<MessageName>::<SignalName>
        """
        return self.sig_com_obj.FullName

    @property
    def is_online(self) -> bool:
        """Checks whether the measurement is running and the signal has been received.

        Returns:
            bool: TRUE: if the measurement is running and the signal has been received. FALSE: if not.
        """
        return self.sig_com_obj.IsOnline

    @property
    def raw_value(self) -> int:
        """Returns the current value of the signal as it was transmitted on the bus.

        Returns:
            int: The raw value of the signal.
        """
        return self.sig_com_obj.RawValue

    @raw_value.setter
    def raw_value(self, value: int) -> None:
        """Returns the current value of the signal as it was transmitted on the bus.

        Returns:
            int: The raw value of the signal.
        """
        self.sig_com_obj.RawValue = value

    @property
    def state(self) -> int:
        """Returns the state of the signal.

        Returns:
            int: State of the signal; possible values are: 0: The default value of the signal is returned. 1: The measurement is not running; the value set by the application is returned. 3: The signal has been received in the current measurement; the current value is returned.
        """
        return self.sig_com_obj.State

    @property
    def value(self) -> int:
        """gets the active value of the signal.


        Returns:
            int: The value of the signal
        """
        return self.sig_com_obj.Value

    @value.setter
    def value(self, value: int) -> None:
        """sets the active value of the signal.

        Args:
            value (int): The new value of the signal.
        """
        self.sig_com_obj.Value = value



# Import Python Libraries here
import win32com.client


class Configuration:
    """The Configuration object represents the active configuration.
    """

    def __init__(self, app_obj) -> None:
        self.app_obj = app_obj
        self.log = self.app_obj.log
        self.conf_com_obj = win32com.client.Dispatch(self.app_obj.app_com_obj.Configuration)
        win32com.client.WithEvents(self.conf_com_obj, CanoeConfigurationEvents)

    @property
    def comment(self) -> str:
        """Gets the comment for the configuration.

        Returns:
            str: The comment.
        """
        return self.conf_com_obj.Comment

    @comment.setter
    def comment(self, text: str) -> None:
        """Defines the comment for the configuration.

        Args:
            text (str): The comment.
        """
        self.conf_com_obj.Comment = text
        self.log.info(f'configuration comment set to {text}.')

    @property
    def fdx_enabled(self) -> int:
        """Enables/Disables value of FDX protocol.

        Returns:
            int: The activation state of the FDX protocol. 0: FDX protocol is deactivated. 1: FDX protocol is activated.
        """
        return self.conf_com_obj.FDXEnabled

    @fdx_enabled.setter
    def fdx_enabled(self, enabled: int) -> None:
        """Enables/Disables the FDX protocol.

        Args:
            enabled (int): The activation state of the FDX protocol. 0: deactivate FDX protocol. ≠0: activate FDX protocol.
        """
        self.conf_com_obj.FDXEnabled = enabled
        self.log.info(f'FDX protocol set to {enabled}.')

    @property
    def full_name(self) -> str:
        """gets the complete path of the configuration.

        Returns:
            str: complete path of the configuration.
        """
        return self.conf_com_obj.FullName

    @full_name.setter
    def full_name(self, full_name: str) -> None:
        """sets the complete path of the configuration.

        Args:
            full_name (str): The new complete path of the configuration.
        """
        self.conf_com_obj.FullName = full_name
        self.log.info(f'complete path of the configuration set to {full_name}.')

    @property
    def mode(self) -> int:
        """returns whether the Online mode or the Offline mode is active.

        Returns:
            int: The currently active mode.
        """
        return self.conf_com_obj.Mode

    @mode.setter
    def mode(self, mode: int) -> None:
        """sets the Online mode or the Offline mode to active.

        Args:
            mode (int): The active mode; valid values are: 0-Online mode is activated. 1-Offline mode is activated.
        """
        self.conf_com_obj.Mode = mode
        self.log.info(f'offline/online mode set to {mode}.')

    @property
    def modified(self) -> bool:
        """Returns information on whether the current configuration was modified since the time it was loaded or created, or sets this property.
        This property determines whether the user is prompted to save when another configuration is loaded.

        Returns:
            bool: The current value of the property.
        """
        return self.conf_com_obj.Modified

    @modified.setter
    def modified(self, modified: bool) -> None:
        """sets Modified property to flase/true.

        Args:
            modified (bool): Value to be assigned to the Modified property.
        """
        self.conf_com_obj.Modified = modified
        self.log.info(f'configuration modified property set to {modified}.')

    @property
    def name(self) -> str:
        """Returns the name of the configuration.

        Returns:
            str: The name of the currently loaded configuration.
        """
        return self.conf_com_obj.Name

    @property
    def path(self) -> str:
        """returns the path of the configuration, depending on the actual configuration.

        Returns:
            str: The path of the currently loaded configuration.
        """
        return self.conf_com_obj.Path

    @property
    def read_only(self) -> bool:
        """Indicates whether the configuration is write protected.

        Returns:
            bool: If the object is write protected True is returned; otherwise False is returned.
        """
        return self.conf_com_obj.ReadOnly

    @property
    def saved(self) -> bool:
        """Indicates whether changes to the configuration have already been saved.

        Returns:
            bool: If changes were made to the configuration and they have not been saved yet, False is returned; otherwise True is returned.
        """
        return self.conf_com_obj.Saved

    def compile_and_verify(self):
        """Compiles all CAPL test modules and verifies all XML test modules.
        All test modules in the Simulation Setup and in the Test Setup are taken into consideration.
        """
        self.conf_com_obj.CompileAndVerify()
        self.log.info(f'Compiled all test modules in the Simulation Setup and in the Test Setup.')

    def save(self, path='', prompt_user=False):
        """Saves the configuration.

        Args:
            path (str): The complete file name. If no path is specified, the configuration is saved under its current name. If it is not saved yet, the user will be prompted for a name.
            prompt_user (bool, optional): A boolean value that indicates whether the user should intervene in error situations.
        """
        if not self.saved:
            if path == '':
                self.conf_com_obj.Save()
            else:
                self.conf_com_obj.Save(path, prompt_user)
            self.log.info(f'Saved configuration({path}).')
        else:
            self.log.info('CANoe Configuration already in saved state.')
        return self.saved

    def save_as(self, path: str, major: str, minor: str, prompt_user: bool):
        """Saves the configuration as a different CANoe version

        Args:
            path (str): The complete path.
            major (str): The major version number of the target version, e.g. 10 for CANoe 10.1.
            minor (str): The minor version number of the target version, e.g. 1 for CANoe 10.1
            prompt_user (bool): A boolean value that defines whether the user should interact in error situations.
        """
        self.conf_com_obj.SaveAs(path, major, minor, prompt_user)
        self.log.info(f'Saved configuration as {path}.')


class CanoeConfigurationEvents:
    """Handler for CANoe Configuration events"""

    @staticmethod
    def OnClose():
        """Occurs when the configuration is closed.
        """
        print('configuration OnClose event triggered.')

    @staticmethod
    def OnSystemVariablesDefinitionChanged():
        """Occurs when system variable definitions are added, changed or removed.
        """
        print('configuration OnSystemVariablesDefinitionChanged event triggered.')



# Import Python Libraries here
import win32com.client


class Version:
    """The Version object represents the version of the CANoe application.
    """

    def __init__(self, app_obj) -> None:
        self.app_obj = app_obj
        self.log = self.app_obj.log
        self.ver_obj = win32com.client.Dispatch(self.app_obj.app_com_obj.Version)

    @property
    def build(self) -> int:
        """Returns the build number of the CANoe application.

        Returns:
            int: The build number of the CANoe application.
        """
        return self.ver_obj.Build

    @property
    def full_name(self) -> str:
        """Determines the complete path of the object.

        Returns:
            str: The complete CANoe version in the following format: "Vector CANoe /run 6.0.50" or "Vector CANoe.LIN /run 6.0.50".
        """
        return self.ver_obj.FullName

    @full_name.setter
    def full_name(self, full_name: str) -> None:
        """Sets the complete path of the object.

        Args:
            full_name (str): The complete CANoe version in the following format: "Vector CANoe /run 6.0.50" or "Vector CANoe.LIN /run 6.0.50".
        """
        self.ver_obj.FullName = full_name
        self.log.info(f'CANoe version set to {full_name}.')

    @property
    def major(self) -> int:
        """Returns the major version number of the CANoe application.

        Returns:
            int: The major version number of the CANoe application.
        """
        return self.ver_obj.major

    @property
    def minor(self) -> int:
        """Returns the Minor version number of the CANoe application.

        Returns:
            int: The Minor version number of the CANoe application.
        """
        return self.ver_obj.minor

    @property
    def name(self) -> str:
        """Returns the name of the object.

        Returns:
            str: The CANoe version in the following format: "CANoe 5.1 SP2" (with Service Pack) or "CANoe.LIN 5.1" (without Service Pack).
        """
        return self.ver_obj.Name

    @property
    def patch(self) -> int:
        """Returns the patch number of the CANoe application.

        Returns:
            int: The patch number of the CANoe application.
        """
        return self.ver_obj.Patch



# Import Python Libraries here
import pythoncom
import win32com.client
from time import sleep as wait


class Measurement:
    """The Measurement object represents measurement functions of CANoe.
    """
    STARTED = False
    STOPPED = False

    def __init__(self, app, user_capl_function_names=tuple()) -> None:
        self.app = app
        self.log = self.app.log
        CanoeMeasurementEvents.app_com_obj = self.app.app_com_obj
        CanoeMeasurementEvents.user_capl_function_names = user_capl_function_names
        self.meas_obj = win32com.client.Dispatch(self.app.app_com_obj.Measurement)
        self.wait_for_canoe_meas_to_start = lambda: DoEventsUntil(lambda: Measurement.STARTED)
        self.wait_for_canoe_meas_to_stop = lambda: DoEventsUntil(lambda: Measurement.STOPPED)
        win32com.client.WithEvents(self.meas_obj, CanoeMeasurementEvents)

    @property
    def animation_delay(self) -> int:
        """Defines the animation delay during the measurement in Offline mode.

        Returns:
            int: The animation delay during the measurement in Offline mode.
        """
        return self.meas_obj.AnimationDelay

    @animation_delay.setter
    def animation_delay(self, delay: int):
        """Sets the animation delay during the measurement in Offline mode.

        Args:
            delay (int): Animation delay
        """
        self.meas_obj.AnimationDelay = delay
        self.log.info(f'Animation delay set to = {delay}.')

    @property
    def measurement_index(self) -> int:
        """Determines the measurement index for the next measurement.

        Returns:
            int: Returns the measurement index for the next measurement.
        """
        return self.meas_obj.MeasurementIndex

    @measurement_index.setter
    def measurement_index(self, index: int):
        """sets the measurement index for the next measurement.

        Args:
            index (int): The measurement index for the next measurement.
        """
        self.meas_obj.MeasurementIndex = index
        self.log.info(f'next measurement index set to = {index}.')

    @property
    def running(self) -> bool:
        """Returns the running state of the measurement.

        Returns:
            bool: True- The measurement is running. False- The measurement is not running.
        """
        return self.meas_obj.Running

    @property
    def user_capl_function_obj_dict(self):
        return CanoeMeasurementEvents.user_capl_function_obj_dict

    def animate(self) -> None:
        """Starts the measurement in Animation mode.
        """
        self.meas_obj.Animate()
        self.log.info(f'Started the measurement in Animation mode with animation delay = {self.animation_delay}.')

    def break_offline_mode(self) -> None:
        """Interrupts the playback in Offline mode.
        """
        if self.running:
            self.meas_obj.Break()
            self.log.info('Interrupted the playback in Offline mode.')

    def reset_offline_mode(self) -> None:
        """Resets the measurement in Offline mode.
        """
        self.meas_obj.Reset()
        self.log.info('resetted measurement in offline mode.')

    def start(self) -> bool:
        """Starts the measurement.
        """
        if not self.running:
            self.meas_obj.Start()
            if not self.running:
                self.log.info(f'waiting for measurement to start...')
                self.wait_for_canoe_meas_to_start()
            self.log.info(f'CANoe Measurement Started. Measurement running status = {self.running}')
        else:
            self.log.info(f'CANoe Measurement Already Running. Measurement running status = {self.running}')
        return self.running

    def step(self) -> None:
        """Processes a measurement event in single step.
        """
        if not self.running:
            self.meas_obj.Step()
            self.log.info('processed a measurement event in single step')

    def stop(self) -> bool:
        """Stops the measurement.
        """
        return self.stop_ex()

    def stop_ex(self) -> bool:
        """StopEx repairs differences in the behavior of the Stop method on deferred stops concerning simulated and real mode in CANoe.
        Calling the StopEx method correlates to clicking the Stop button.
        """
        if self.running:
            self.meas_obj.StopEx()
            if self.running:
                self.log.info(f'waiting for measurement to stop...')
                self.wait_for_canoe_meas_to_stop()
            self.log.info(f'CANoe Measurement Stopped. Measurement running status = {self.running}')
        else:
            self.log.info(f'CANoe Measurement Already Stopped. Measurement running status = {self.running}')
        return not self.running


class CanoeMeasurementEvents:
    """Handler for CANoe Measurement events"""
    app_com_obj = object
    user_capl_function_names = tuple()
    user_capl_function_obj_dict = dict()

    @staticmethod
    def OnInit():
        """Occurs when the measurement is initialized.
        """
        app_com_obj_loc = CanoeMeasurementEvents.app_com_obj
        for fun in CanoeMeasurementEvents.user_capl_function_names:
            CanoeMeasurementEvents.user_capl_function_obj_dict[fun] = app_com_obj_loc.CAPL.GetFunction(fun)
        print('measurement OnInit event triggered')

    @staticmethod
    def OnExit():
        """Occurs when the measurement is exited.
        """
        print('measurement OnExit event triggered')

    @staticmethod
    def OnStart():
        """Occurs when the measurement is started.
        """
        Measurement.STARTED = True
        Measurement.STOPPED = False
        print('measurement OnStart event triggered')

    @staticmethod
    def OnStop():
        """Occurs when the measurement is stopped.
        """
        Measurement.STARTED = False
        Measurement.STOPPED = True
        print('measurement OnStop event triggered')


def DoEvents():
    pythoncom.PumpWaitingMessages()
    wait(.1)


def DoEventsUntil(cond):
    while not cond():
        DoEvents()



# Import Python Libraries here
import os
import sys
import logging
from logging import handlers


class PyCanoeLogger:
    def __init__(self, py_canoe_log_dir='') -> None:
        self.log = logging.getLogger('CANOE_LOG')
        self.log.propagate = False
        self.__py_canoe_log_initialisation(py_canoe_log_dir)

    def __py_canoe_log_initialisation(self, py_canoe_log_dir):
        self.log.setLevel(logging.DEBUG)
        log_format = logging.Formatter("%(asctime)s [CANOE_LOG] [%(levelname)-5.5s] %(message)s")
        ch = logging.StreamHandler(sys.stdout)
        ch.setFormatter(log_format)
        self.log.addHandler(ch)
        if py_canoe_log_dir != '' and not os.path.exists(py_canoe_log_dir):
            os.makedirs(py_canoe_log_dir, exist_ok=True)
        if os.path.exists(py_canoe_log_dir):
            fh = handlers.RotatingFileHandler(fr'{py_canoe_log_dir}\py_canoe.log', maxBytes=(1024 * 50), backupCount=20)
            fh.setFormatter(log_format)
            self.log.addHandler(fh)



# Import Python Libraries here
import win32com.client


class Capl:
    """The CAPL object allows to compile all nodes (CAPL, .NET, XML) in the configuration.
    Additionally, it represents the CAPL functions available in the CAPL programs.
    Please note that only user-defined CAPL functions can be accessed
    """

    def __init__(self, app_obj: object) -> None:
        self.app_obj = app_obj
        self.log = self.app_obj.log
        self.capl_obj = win32com.client.Dispatch(self.app_obj.app_com_obj.CAPL)

    def compile(self) -> None:
        """Translates all CAPL, XML and .NET nodes.
        """
        self.capl_obj.Compile()

    def get_function(self, name: str) -> object:
        """Returns a CAPLFunction object.

        Args:
            name (str): The name of the CAPL function.

        Returns:
            object: The CAPLFunction object.
        """
        return self.capl_obj.GetFunction(name)

    @staticmethod
    def parameter_count(capl_function_object: object) -> int:
        """Returns the number of parameters of the CAPL function.

        Args:
            capl_function_object (object): The CAPLFunction object.

        Returns:
            int: The number of parameters of the CAPL function.
        """
        return capl_function_object.ParameterCount

    @staticmethod
    def parameter_types(capl_function_object: object) -> tuple:
        """Returns the types of the parameters of the CAPL function as byte array.
        The parameter types are coded as follows:
        L: long (32 bit signed integer)
        D: dword (32 bit unsigned integer)
        F: double (64 bit floating point)

        Args:
            capl_function_object (object): The CAPLFunction object.

        Returns:
            tuple: The types of the parameters of the CAPL function as byte array.
        """
        return capl_function_object.ParameterTypes

    def call_capl_function(self, capl_function_obj: object, *arguments) -> bool:
        """Calls a CAPL function.
        Please note that the number of parameters must agree with that of the CAPL function.
        The return value is only available for CAPL functions whose CAPL programs are configured in the Measurement Setup.
        Only integers are allowed as a return type.

        Args:
            capl_function_obj (str): CAPL function object.
            arguments (tuple): Function parameters p1…p10 (optional).

        Returns:
            bool: CAPL function execution status. True-success, False-failed.
        """
        return_value = False
        if len(arguments) == self.parameter_count(capl_function_obj):
            if len(arguments) > 0:
                capl_function_obj.Call(*arguments)
            else:
                capl_function_obj.Call()
            return_value = True
        else:
            print(fr'function arguments not matching with CAPL user function args.')
        return return_value

    def compile_result(self) -> dict:
        """The CompileResult object represents the result of the last compilation of the CAPL object.

        Returns:
            dict: returns dictionary of 'error_message', 'node_name', 'result', 'source_file'
        """
        return_values = dict()
        compile_result_obj = self.capl_obj.CompileResult
        # Returns the last compilation error for the CompileResult object or the last loading error/warning for the OpenConfigurationResult object
        return_values['error_message'] = compile_result_obj.ErrorMessage
        # Returns the name of the first compilation error node.
        return_values['node_name'] = compile_result_obj.NodeName
        # Returns the result of the last compilation of the CAPL object.
        return_values['result'] = compile_result_obj.Result
        # Returns the path of the program file where the first compile error occurred
        return_values['source_file'] = compile_result_obj.SourceFile
        return return_values



# Import Python Libraries here
import win32com.client


class System:
    def __init__(self, app_obj) -> None:
        self.app_obj = app_obj
        self.log = self.app_obj.log
        self.sys_com_obj = win32com.client.Dispatch(self.app_obj.app_com_obj.System)


class VariablesFiles:
    def __init__(self) -> None:
        pass


class VariablesFile:
    def __init__(self) -> None:
        pass


class Namespaces:
    def __init__(self, namespaces_com_obj: object) -> None:
        self.namespaces_com_obj = namespaces_com_obj

    @property
    def count(self) -> int:
        return self.namespaces_com_obj.Count

    def add(self, name: str) -> object:
        """Adds a new namespace.

        Args:
            name (str): The name of the new namespace.

        Returns:
            object: The new Namespace object.
        """
        return self.namespaces_com_obj.Add(name)

    def remove(self, name: str) -> None:
        """Removes an Namespace from a group

        Args:
            name (str): A Namespace object.
        """
        self.namespaces_com_obj.Remove(name)


class Namespace:
    def __init__(self, namespace_com_obj: object) -> None:
        self.namespace_com_obj = namespace_com_obj

    @property
    def comment(self) -> str:
        """Gets the comment for the Namespace.

        Returns:
            str: The comment.
        """
        return self.namespace_com_obj.Comment

    @comment.setter
    def comment(self, text: str) -> None:
        """Defines the comment for the Namespace.

        Args:
            text (str): The comment
        """
        self.namespace_com_obj.Comment = text

    @property
    def name(self) -> str:
        """Returns the name of the Namespace.

        Returns:
            str: The name of the namespace.
        """
        return self.namespace_com_obj.Name

    @property
    def namespaces(self) -> object:
        """Returns the Namespaces object.

        Returns:
            object: The Namespaces object.
        """
        return self.namespace_com_obj.Namespaces

    @property
    def variables(self) -> object:
        """Returns the Variables object.

        Returns:
            object: The Variables object.
        """
        return self.namespace_com_obj.Variables


class Variables:
    def __init__(self, variables_com_obj) -> None:
        self.variables_com_obj = variables_com_obj

    @property
    def count(self) -> int:
        """Returns the number of Variable objects inside the collection.

        Returns:
            int: _description_
        """
        return self.variables_com_obj.Count

    def add(self, name: str, initial_value=0) -> object:
        """Adds a new read-only variable.

        Args:
            name (str): The name of the new variable.
            initial_value (int): The initial value of the new variable. default value: 0 (Integer).

        Returns:
            object: The new Variable object.
        """
        return self.variables_com_obj.Add(name, initial_value)

    def add_ex(self, name: str, initial_value=0, min_value=0, max_value=0) -> object:
        """Adds a new read-only variable.

        Args:
            name (str): The name of the new variable.
            initial_value (int, optional): The initial value of the new variable. Defaults to 0.
            min_value (int, optional): The minimum value of the new variable. Defaults to 0.
            max_value (int, optional): The maximum value of the new variable. Defaults to 0.

        Returns:
            object: The new Variable object.
        """
        return self.variables_com_obj.AddEx(name, initial_value, min_value, max_value)

    def add_writable(self, name: str, initial_value=0) -> object:
        """Adds a new writable variable.

        Args:
            name (str): The name of the new variable.
            initial_value (int): The initial value of the new variable. default value: 0 (Integer).

        Returns:
            object: The new Variable object.
        """
        return self.variables_com_obj.AddWriteable(name, initial_value)

    def add_writable_ex(self, name: str, initial_value=0, min_value=0, max_value=0) -> object:
        """Adds a new writable variable.

        Args:
            name (str): The name of the new variable.
            initial_value (int, optional): The initial value of the new variable. Defaults to 0.
            min_value (int, optional): The minimum value of the new variable. Defaults to 0.
            max_value (int, optional): The maximum value of the new variable. Defaults to 0.

        Returns:
            object: The new Variable object.
        """
        return self.variables_com_obj.AddWritableEx(name, initial_value, min_value, max_value)

    def remove(self, variable: object) -> None:
        """Removes variable from a group

        Args:
            variable (str): Variable object.
        """
        self.variables_com_obj.Remove(variable)


class Variable:
    def __init__(self, variable_com_obj) -> None:
        self.variable_com_obj = variable_com_obj

    @property
    def analysis_only(self) -> bool:
        """Determines if the variable shall be only used for analysis purposes or not.

        Returns:
            bool: false (default) ,true
        """
        return self.variable_com_obj.AnalysisOnly

    @analysis_only.setter
    def analysis_only(self, value: bool) -> None:
        """sets if the variable shall be only used for analysis purposes or not.
        If the property is set to false (default value), it may still be changed to analysis only in a CAPL program.

        Args:
            value (bool): false (default) ,true.
        """
        self.variable_com_obj.AnalysisOnly = value

    @property
    def bit_count(self) -> int:
        """Returns the number of bits of the variable data type.

        Returns:
            int: The number of bits of the variable data type.
        """
        return self.variable_com_obj.BitCount

    @property
    def comment(self) -> str:
        """Gets the comment for the variable..

        Returns:
            str: The comment.
        """
        return self.variable_com_obj.Comment

    @comment.setter
    def comment(self, text: str) -> None:
        """defines the comment for the variable.

        Args:
            text (str): The comment
        """
        self.variable_com_obj.Comment = text

    @property
    def element_count(self) -> int:
        """For arrays: the maximum number of elements in the array.

        Returns:
            int: The maximum number of elements in the array.
        """
        return self.variable_com_obj.ElementCount

    @property
    def full_name(self) -> str:
        """determines the complete path of the variable.

        Returns:
            str: The full name, including namespace, variable name and member name.
        """
        return self.variable_com_obj.FullName

    @full_name.setter
    def full_name(self, full_name: str) -> None:
        """sets the complete path of the variable.

        Args:
            full_name (str): The new complete path of the object.
        """
        self.variable_com_obj.FullName = full_name

    @property
    def name(self) -> str:
        """Returns the name of the variable.

        Returns:
            str: The name of the system variable.
        """
        return self.variable_com_obj.Name

    @property
    def init_value(self) -> tuple[int, float, str]:
        """The initial value of the variable.

        Returns:
            tuple[int, float, str]: The initial value of the variable.
        """
        return self.variable_com_obj.InitValue

    @property
    def min_value(self) -> tuple[int, float, str]:
        """Returns the minimum value of the object.

        Returns:
            tuple[int, float, str]: minimum value of the variable.
        """
        return self.variable_com_obj.MinValue

    @property
    def max_value(self) -> tuple[int, float, str]:
        """Returns the maximum value of the variable.

        Returns:
            tuple[int, float, str]: The maximum value of the variable.
        """
        return self.variable_com_obj.MaxValue

    @property
    def is_array(self) -> bool:
        """Whether the variable data type is an array.

        Returns:
            bool: Whether the variable data type is an array.
        """
        return self.variable_com_obj.IsArray

    @property
    def is_signed(self) -> bool:
        """For integer variables: whether the data type is signed.

        Returns:
            bool: Whether the data type is signed.
        """
        return self.variable_com_obj.IsSigned

    @property
    def read_only(self) -> bool:
        """Indicates whether the system variable is write protected.

        Returns:
            bool: If the variable is write protected True is returned; otherwise False is returned.
        """
        return self.variable_com_obj.ReadOnly

    @property
    def type(self) -> int:
        """Returns the type of a system variable.

        Returns:
            int: The type of the system variable. The following types are define- 0: Integer 1: Float 2: String 4: Float Array 5: Integer Array 6: LongLong 7: Byte Array 98: Generic Array 99: Struct 65535: Invalid
        """
        return self.variable_com_obj.Type

    @property
    def unit(self) -> str:
        """Returns the unit of the variable.

        Returns:
            str: The unit of the variable.
        """
        return self.variable_com_obj.Unit

    @property
    def value(self) -> tuple[int, float, str]:
        """Defines or sets the active value of the variable.

        Returns:
            tuple[int, float, str]: The value of the variable.
        """
        return self.variable_com_obj.Value

    @value.setter
    def value(self, value: tuple[int, float, str]) -> None:
        """Defines or sets the active value of the variable.

        Args:
            value (tuple[int, float, str]): The new value of the variable.
        """
        self.variable_com_obj.Value = value


class Encodings:
    def __init__(self) -> None:
        pass


class Encoding:
    def __init__(self) -> None:
        pass



# Import Python Libraries here
import os


class Application:
    """The Application object represents the CANoe application.
    """
    OPENED = False
    CLOSED = False

    def __init__(self, log_obj) -> None:
        self.app_com_obj = None
        self.log = log_obj

    @property
    def channel_mapping_name(self) -> str:
        """get the application name which is used to map application channels to real existing Vector hardware interface channels.

        Returns:
            str: The application name
        """
        return self.app_com_obj.ChannelMappingName

    @channel_mapping_name.setter
    def channel_mapping_name(self, name: str):
        """set the application name which is used to map application channels to real existing Vector hardware interface channels.

        Args:
            name (str): The application name used to map the channels.
        """
        self.app_com_obj.ChannelMappingName = name

    @property
    def full_name(self) -> str:
        """determines the complete path of the CANoe application.

        Returns:
            str: location where CANoe is installed.
        """
        return self.app_com_obj.FullName

    @property
    def name(self) -> str:
        """Returns the name of the CANoe application.

        Returns:
            str: name of the CANoe application.
        """
        return self.app_com_obj.Name

    @property
    def path(self) -> str:
        """Returns the Path of the CANoe application.

        Returns:
            str: Path of the CANoe application.
        """
        return self.app_com_obj.Path

    @property
    def visible(self) -> bool:
        """Returns whether the CANoe main window is visible or is only displayed by a tray icon.

        Returns:
            bool: A boolean value indicating whether the CANoe main window is visible..
        """
        return self.app_com_obj.Visible

    @visible.setter
    def visible(self, visible: bool):
        """Defines whether the CANoe main window is visible or is only displayed by a tray icon.

        Args:
            visible (bool): A boolean value indicating whether the CANoe main window is to be visible.
        """
        self.app_com_obj.Visible = visible

    def new(self, auto_save=False, prompt_user=False) -> None:
        """Creates a new configuration.

        Args:
            auto_save (bool, optional): A boolean value that indicates whether the active configuration should be saved if it has been changed. Defaults to False.
            prompt_user (bool, optional): A boolean value that indicates whether the user should intervene in error situations. Defaults to False.
        """
        self.app_com_obj.New(auto_save, prompt_user)
        self.log.info('created a new configuration...')

    def open(self, path: str, auto_save=False, prompt_user=False) -> None:
        """Loads a configuration.

        Args:
            path (str): The complete path for the configuration.
            auto_save (bool, optional): A boolean value that indicates whether the active configuration should be saved if it has been changed. Defaults to False.
            prompt_user (bool, optional): A boolean value that indicates whether the user should intervene in error situations. Defaults to False.

        Raises:
            FileNotFoundError: _description_
        """
        if os.path.isfile(path):
            self.log.info(f'CANoe cfg "{path}" found.')
            self.app_com_obj.Open(path, auto_save, prompt_user)
            self.log.info(f'loaded CANoe config "{path}"')
        else:
            self.log.info(f'CANoe cfg "{path}" not found.')
            raise FileNotFoundError(f'CANoe cfg file "{path}" not found!')

    def quit(self):
        """Quits the application.
        """
        self.app_com_obj.Quit()
        self.log.info('CANoe Application Closed.')


class ApplicationEvents:
    """Handler for CANoe Application events"""

    @staticmethod
    def OnOpen():
        """Occurs when a configuration is opened.
        """
        print('canoe opened')
        Application.OPENED = True
        Application.CLOSED = False

    @staticmethod
    def OnQuit():
        """Occurs when CANoe is quit
        """
        print('canoe closed')
        Application.OPENED = False
        Application.CLOSED = True
